"""
License
-------
    The MIT License (MIT)

    Copyright (c) 2018 Snappy2 Project

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

Created on Sat Dec 16 22:46:28 2017

@author: Ahmad Barqawi
@Source: Github.com/Barqawiz
"""
class Constants:
    # draw based on position
    pos_eyes = 'eyes'
    pos_ears = 'ears'
    pos_head = 'head'
    pos_must = 'mustache'
    pos_noise = 'noise'
    pos_lips = 'lips'


class Overlays:
    # draw based on objects
    draw_mus = ['mus1.png', 'mus2.png', 'mus3.png']
    draw_glasses = ['sunglasses1.png', 'sunglasses2.png']
    draw_hat = ['hat1.png', 'hat2.png']
    draw_ears = ['ears_bear.png', 'ears_cat.png', 'ears_teady.png']



"""
License
-------
    The MIT License (MIT)

    Copyright (c) 2018 Snappy2 Project

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

Created on Sat Dec 16 22:46:28 2017

@author: Ahmad Barqawi
@Source: Github.com/Barqawiz
"""
from snappy2.utils import Constants
from snappy2.utils.Utility import Utility
import PIL.Image as Image


class FacePainter:
    """
    Draw overlay images above faces using key points
        - draw overlay
        - draw glasses
        - draw mustache
        - draw hat
        - draw ears
        - draw key points


    Developed by: github.com/barqawiz
    """
    def __init__(self, image, key_properties):
        self.image = image
        self.key_properties = key_properties

    def draw_overlay(self, img_overlay, position=Constants.pos_eyes, x_margin_m=25,
                        y_margin_m=10, x_slide_m=0, y_slide_m=0, width_increase_p=0, extra_angle_m=0):
        """
        Draw overlay image above the detected face based on the received position
        :param img_overlay:
        :param position:
        :param x_margin_m: if the face is big, this parameter slide the overlay based on value and increase size of image
        :param y_margin_m: slide the overlay to up
        :param x_slide_m: extra slide for x positive or negative without increase size
        :param y_slide_m: extra slide for y positive or negative
        :param width_increase_p: percentage of increase in the width of overlay (smi supported in this version)
        :param extra_angle_m add extra value to rotation, positive or negative
        :return:
        """

        x_margin_m = abs(x_margin_m)
        for face in self.key_properties:
            x_axis = face['keys_x']
            y_axis = face['keys_y']
            face_y = face['face_y']
            face_w = face['face_w']
            face_h = face['face_w']
            # reset params
            x_margin = x_margin_m
            y_margin = y_margin_m
            x_slide = x_slide_m
            y_slide = y_slide_m
            extra_angle = extra_angle_m

            sx = None
            sy = None
            if position == Constants.pos_eyes:

                # common fixing values
                # space_to_eye = y_axis[9] - face_y
                x_margin += int((x_axis[6] - x_axis[8]) * 0.33)

                # size change
                target_width = int(x_axis[7] - x_axis[9]) + x_margin  # + extra margin (left and right)

                percent_increase = int(target_width * width_increase_p)
                target_width += percent_increase

                overlay_width = int(img_overlay.size[0])
                change_ratio = (target_width / overlay_width)

                target_width = int(img_overlay.size[0] * change_ratio)
                target_height = int(img_overlay.size[1] * change_ratio)

                # angle change
                angle = Utility.calc_angle(x_axis, y_axis) + extra_angle

                # prepare draw sun glasses params
                draw_height, draw_width = img_overlay.size
                sm = int(x_margin / 2) + int(percent_increase/2)
                sx = int(x_axis[9]) - sm
                sy = int(y_axis[9]) - y_margin
                sx += x_slide

                # transform overlay image
                img_overlay = img_overlay.resize((target_width, target_height), Image.ANTIALIAS)
                img_overlay = img_overlay.rotate(angle, resample=Image.BICUBIC, expand=True)

                if angle > 4:
                    if abs(y_axis[9] - y_axis[7]) > 10:
                        # slide a little
                        sx = sx - sm

            elif position == Constants.pos_must:

                # common fixing values
                space_to_mouth = y_axis[13] - y_axis[10]
                # print(space_to_mouth)
                if space_to_mouth >= 36:
                    y_margin += int(space_to_mouth * 0.2)
                elif space_to_mouth >= 32:
                    y_margin += int(space_to_mouth * 0.1)
                else:
                    y_margin += int(space_to_mouth * 0.05)
                #
                # size change
                target_width = int(x_axis[11] - x_axis[12]) + x_margin  # + extra margin (left and right)
                overlay_width = int(img_overlay.size[0])
                change_ratio = (target_width / overlay_width)

                target_width = int(img_overlay.size[0] * change_ratio)
                target_height = int(img_overlay.size[1] * change_ratio)

                # angle change
                angle = Utility.calc_angle(x_axis, y_axis) + extra_angle

                # prepare draw sun glasses params
                draw_height, draw_width = img_overlay.size
                sm = int(x_margin / 2)
                sx = int(x_axis[10]) - int(target_width/2) + x_slide
                sy = int(y_axis[10]) + y_margin

                # transform overlay image
                img_overlay = img_overlay.resize((target_width, target_height), Image.ANTIALIAS)
                img_overlay = img_overlay.rotate(angle, resample=Image.BICUBIC, expand=True)

            elif position == Constants.pos_ears:

                # common fixing values
                # space_to_eye = y_axis[9] - face_y
                x_margin += int(x_axis[6] - x_axis[8])

                # size change
                target_width = int(x_axis[7] - x_axis[9]) + x_margin  # + extra margin (left and right)
                overlay_width = int(img_overlay.size[0])
                change_ratio = (target_width / overlay_width)

                target_width = int(img_overlay.size[0] * change_ratio)
                target_height = int(img_overlay.size[1] * change_ratio)
                y_margin += int(target_height * 0.85)

                # angle change
                angle = Utility.calc_angle(x_axis, y_axis) + extra_angle

                # prepare draw sun glasses params
                draw_height, draw_width = img_overlay.size
                sm = int(x_margin / 2)
                sx = int(x_axis[9]) - sm
                sy = int(face_y) - y_margin
                sx += x_slide

                # transform overlay image
                img_overlay = img_overlay.resize((target_width, target_height), Image.ANTIALIAS)
                img_overlay = img_overlay.rotate(angle, resample=Image.BICUBIC, expand=True)

            elif position == Constants.pos_head:
                # common fixing values
                space_to_eye = int( (y_axis[9] - face_y)*0.3)
                x_margin += int(x_axis[6] - x_axis[8])*2

                # print(space_to_eye)

                # size change
                target_width = int(x_axis[7] - x_axis[9]) + x_margin  # + extra margin (left and right)
                overlay_width = int(img_overlay.size[0])
                change_ratio = (target_width / overlay_width)

                target_width = int(img_overlay.size[0] * change_ratio)
                target_height = int(img_overlay.size[1] * change_ratio)
                y_margin += int(target_height * 0.7)
                y_margin += int(space_to_eye * 0.9)

                # angle change
                angle = Utility.calc_angle(x_axis, y_axis) + extra_angle

                # prepare draw sun glasses params
                draw_height, draw_width = img_overlay.size
                sm = int(x_margin / 2)
                sx = int(x_axis[9]) - sm
                sy = int(face_y) - y_margin + y_slide
                sx += x_slide

                # transform overlay image
                img_overlay = img_overlay.resize((target_width, target_height), Image.ANTIALIAS)
                img_overlay = img_overlay.rotate(angle, resample=Image.BICUBIC, expand=True)
            else:
                raise Exception('unsupported position, check constants file for supported values')

            # draw overlay image
            if sx is not None and sy is not None:
                self.image.paste(img_overlay, (sx, sy), img_overlay)

    def draw_glasses(self, img_overlay, x_margin=5, y_margin=8, x_slide=0):
        """
        Draw overlay image on top the detected eyes

        :param img_overlay:
        :param x_margin: if the face is big, this parameter slide the overlay based on value and increase size of image
        :param y_margin: slide the overlay to up
        :param x_slide: extra slide for x positive or negative without increase size
        :return:
        """
        self.draw_overlay(img_overlay, position=Constants.pos_eyes, x_margin_m=x_margin, y_margin_m=y_margin,
                          x_slide_m=x_slide)

    def draw_mustache(self, img_overlay, x_margin=5, y_margin=5, x_slide=0):
        """
        Draw overlay image above the detected mouth

        :param img_overlay:
        :param x_margin: if the face is big, this parameter slide the overlay based on value and increase size of image
        :param y_margin: slide the overlay to up
        :param x_slide: extra slide for x positive or negative without increase size
        :return:
        """
        self.draw_overlay(img_overlay, position=Constants.pos_must, x_margin_m=x_margin, y_margin_m=y_margin,
                          x_slide_m=x_slide)

    def draw_hat(self, img_overlay, x_margin=25, y_margin=5, x_slide=-11, y_slide=0, width_increase_p=0, extra_angle=0):
        self.draw_overlay(img_overlay, position=Constants.pos_head, x_margin_m=x_margin, y_margin_m=y_margin,
                          x_slide_m=x_slide, y_slide_m=y_slide, width_increase_p=width_increase_p, extra_angle_m=extra_angle)

    def draw_ears(self, img_overlay, x_margin=4, y_margin=4, x_slide=0):
        self.draw_overlay(img_overlay, position=Constants.pos_ears, x_margin_m=x_margin, y_margin_m=y_margin,
                          x_slide_m=x_slide)

    def draw_key_points(self):
        if self.image is not None:
            print('not supported for this version, will be released in next versions')



from setuptools import setup, find_packages
import os

with open("PIPREADME.md", "r") as fh:
    long_description = fh.read()

def get_resources():
    resources = []
    for root, dirs, files in os.walk('snappy2/resource'):
        resources.extend([os.path.join(root, f) for f in files])
    return resources

resources = get_resources()
# print(resources)

setup(
    name='snappy2',
    version='0.2',
    author='Albarqawi',
    packages=find_packages(),
    include_package_data=True,
    url='https://snappy2.ahmadai.com/',
    description="Detect faces and draw overlay images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[ 'opencv-python==3.4.18.65', 'Pillow==9.4.0', 'tensorflow==2.9.0', 'keras==2.9.0', 'numpy==1.24.1', 'Markdown==3.4.1', 'absl-py==1.4.0', 'Werkzeug==2.2.2', 'h5py==3.1.0'    ],
    data_files=[('snappy2/resource', resources)]
)



"""
License
-------
    The MIT License (MIT)

    Copyright (c) 2018 Snappy2 Project

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

Created on Sat Dec 16 22:46:28 2017

@author: Ahmad Barqawi
@Source: Github.com/Barqawiz
"""

import cv2
from tensorflow.compat.v1.keras.models import load_model
import numpy as np
from snappy2.utils.Utility import Utility
import os


class Detector:
    """
    Common class to detect areas of interest from images

    Developed by: github.com/barqawiz
    """
    def __init__(self):
        base_folder = os.path.dirname(__file__)
        self.face_cascade = cv2.CascadeClassifier(
            os.path.join(base_folder,'../resource/haarcascades/haarcascade_frontalface_default.xml'))
        self.model = load_model(os.path.join(base_folder,  '../resource/models/keras_cv_base_model_1_avg.h5'))

    def detect_faces(self, gray_human_image):
        """
        Detect faces in image and return list of faces with related properties
        :param gray_human_image: numpy gray image
        :return: face_properties a list of dictionary with following information (face, loc, size)
        """
        face_properties = []
        faces = self.face_cascade.detectMultiScale(gray_human_image, 1.2, 7, minSize=(30, 30))
        if len(faces) == 0:
            faces = self.face_cascade.detectMultiScale(gray_human_image, 1.1, 8, minSize=(30, 30))

        for (x, y, w, h) in faces:
            # 1- detect face area
            roi_image = gray_human_image[y:y + h, x:x + w]
            face_properties.append({'face': roi_image, 'loc': (x, y), 'size': (w, h)})

        return face_properties

    def detect_key_points(self, face_properties, network_in_size=96):
        """
        Detect key points areas in the face using neural network.
            - key points for eyes, noise and mouth

        :param face_properties:
        :param network_in_size:
        :return: key_properties a list of dictionary with following information (keys_x, keys_y, face_index)
        """

        key_properties = []
        index = 0
        for face in face_properties:
            # 0- get face information
            roi_image = face['face']
            x,y = face['loc']
            w, h = face['size']

            # 1- pre process
            # prepare face image
            roi_image = cv2.resize(roi_image, (network_in_size, network_in_size))
            # scale pixel values to [0, 1]
            roi_image = roi_image / 255
            # reshape for netowork format
            roi_image = roi_image.reshape(roi_image.shape[0], roi_image.shape[1], 1)
            roi_image = np.array([roi_image])

            # 2- predict face key points
            y_predict = self.model.predict(roi_image)[0]
            # map to original coordinates values
            y_predict = Utility.reverse_nn_normalization(y_predict)

            # 3- extract information
            # get value within the original image
            x_axis = y_predict[0::2]
            y_axis = y_predict[1::2]
            x_scale_factor = (w / 96)
            y_scale_factor = (h / 96)
            x_axis = x_axis * x_scale_factor + x
            y_axis = y_axis * y_scale_factor + y

            key_properties.append({'keys_x': x_axis, 'keys_y': y_axis, 'face_x': x, 'face_y': y,
                                   'face_w': w, 'face_h': h, 'face_index': index})
            index += 1

        return key_properties



"""
License
-------
    The MIT License (MIT)

    Copyright (c) 2018 Snappy2 Project

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

Created on Sat Dec 16 22:46:28 2017

@author: Ahmad Barqawi
@Source: Github.com/Barqawiz
"""
import sys
sys.path.insert(0, '..')

# from utils.Constants import Constants, Overlays
from snappy2.utils import Constants, Overlays
from snappy2.controls import Detector
from snappy2.controls import FacePainter
import PIL.Image as Image
from PIL import ImageFont, ImageDraw
import cv2
import numpy as np
import os


class Snappy2:

    def __init__(self):
        self.detector = Detector()
        self.base_folder = os.path.dirname(__file__)

    def load_image(self, image_path):
        return Image.open(image_path)

    def draw_overlay_path(self, path_human, path_overlay, draw_pos=Constants.pos_eyes):#
        """
        Detect faces in photos and draw overlay images on it

        :param path_human: image file path
        :param path_overlay: image file path
        :param draw_pos: send one of following eyes | ears | head | mustache
        :return:
        """
        img_human = Image.open(path_human)
        img_overlay = Image.open(path_overlay)

        return self.draw_overlay(img_human, img_overlay, draw_pos=draw_pos)

    def draw_overlay(self, img_human, img_overlay, draw_pos=Constants.pos_eyes):
        """
        Detect faces in photos and draw overlay images on it

        :param img_human: PIL Image
        :param img_overlay: PIL Image
        :param draw_pos: send one of following eyes | ears | head | mustache
        :return:
        """
        image_gray = cv2.cvtColor(np.array(img_human), cv2.COLOR_BGR2GRAY)
        face_properties = self.detector.detect_faces(image_gray)
        key_properties = self.detector.detect_key_points(face_properties)

        face_painter = FacePainter(img_human, key_properties)

        if draw_pos == Constants.pos_eyes:
            face_painter.draw_glasses(img_overlay)
        elif draw_pos == Constants.pos_must:
            face_painter.draw_mustache(img_overlay)
        elif draw_pos == Constants.pos_ears:
            face_painter.draw_ears(img_overlay)
        elif draw_pos == Constants.pos_head:
            face_painter.draw_hat(img_overlay)
        elif draw_pos == Constants.pos_noise:
            print('not supported in this version')
        elif draw_pos == Constants.pos_lips:
            print('not supported in this version')

    def set_glasses(self, img_human, gls_index=0):
        """
        draw glasses on face

        :param img_human:
        :param gls_index: position for value inside (utils.Constants.Overlays.draw_glasses) array
        :return:
        """
        assert gls_index < len(Overlays.draw_glasses), 'Error mustache position'
        overlay_name = Overlays.draw_glasses[gls_index]

        file_overlay = os.path.join(self.base_folder, 'resource/images/overlay/' + overlay_name)

        assert os.path.exists(file_overlay), 'Overlay image does not exist'
        img_overlay = Image.open(file_overlay).convert("RGBA")

        image_gray = cv2.cvtColor(np.array(img_human), cv2.COLOR_BGR2GRAY)
        face_properties = self.detector.detect_faces(image_gray)
        key_properties = self.detector.detect_key_points(face_properties)

        face_painter = FacePainter(img_human, key_properties)
        face_painter.draw_glasses(img_overlay)

    def set_mustache(self, img_human, mus_index=0):
        """
        draw mustache on face

        :param img_human:
        :param mus_index: position for value inside (utils.Constants.Overlays.draw_mus) array
        :return:
        """
        assert mus_index < len(Overlays.draw_mus), 'Error mustache position'
        overlay_name = Overlays.draw_mus[mus_index]

        file_overlay = os.path.join(self.base_folder, 'resource/images/overlay/'+overlay_name)

        assert os.path.exists(file_overlay), 'Overlay image does not exist'
        img_overlay = Image.open(file_overlay).convert("RGBA")

        image_gray = cv2.cvtColor(np.array(img_human), cv2.COLOR_BGR2GRAY)
        face_properties = self.detector.detect_faces(image_gray)
        key_properties = self.detector.detect_key_points(face_properties)

        face_painter = FacePainter(img_human, key_properties)
        if mus_index == 1 or mus_index == 2:
            face_painter.draw_mustache(img_overlay, x_margin=30, x_slide=-3)
        else:
            face_painter.draw_mustache(img_overlay)

    def set_hat(self, img_human, hat_index=0, y_slide=0, width_increase_p=0):
        """
        draw hat on head

        :param img_human:
        :param hat_index:
        :param y_slide: slide the hat up or down based on the value
        :param width_increase_p percentage of increase in overlay width
        :return:
        """
        assert hat_index < len(Overlays.draw_hat), 'Error mustache position'
        overlay_name = Overlays.draw_hat[hat_index]

        file_overlay = os.path.join(self.base_folder, 'resource/images/overlay/' + overlay_name)

        assert os.path.exists(file_overlay), 'Overlay image does not exist'
        img_overlay = Image.open(file_overlay).convert("RGBA")

        image_gray = cv2.cvtColor(np.array(img_human), cv2.COLOR_BGR2GRAY)
        face_properties = self.detector.detect_faces(image_gray)
        key_properties = self.detector.detect_key_points(face_properties)

        face_painter = FacePainter(img_human, key_properties)
        if hat_index == 1:
            face_painter.draw_hat(img_overlay, x_slide=-24, y_slide=y_slide, width_increase_p=width_increase_p)
        else:
            face_painter.draw_hat(img_overlay, y_slide=y_slide, width_increase_p=width_increase_p)

    def set_ears(self, img_human, ea_index=0):
        """
        draw ears on head
        :param img_human:
        :param ea_index:
        :return:
        """

        assert ea_index < len(Overlays.draw_ears), 'Error mustache position'
        overlay_name = Overlays.draw_ears[ea_index]

        file_overlay = os.path.join(self.base_folder, 'resource/images/overlay/' + overlay_name)

        assert os.path.exists(file_overlay), 'Overlay image does not exist'
        img_overlay = Image.open(file_overlay).convert("RGBA")

        image_gray = cv2.cvtColor(np.array(img_human), cv2.COLOR_BGR2GRAY)
        face_properties = self.detector.detect_faces(image_gray)
        key_properties = self.detector.detect_key_points(face_properties)

        face_painter = FacePainter(img_human, key_properties)
        face_painter.draw_ears(img_overlay)

    def set_text(self, object_image, draw_text, text_size, text_fill='#FFFFFF',
                 v_position='bottom', h_position='center'):
        """
        draw text of received image

        :param object_image:
        :param draw_text:
        :param text_size:
        :param text_fill: HTML color code
        :param v_position: bottom OR top
        :param h_position: left OR right OR center
        :return:
        """
        assert v_position in ['bottom','top', 'center'], "received position value not supported." +\
                                                         "\nValue should be bottom or top"
        assert h_position in ['center', 'left', 'right'], "received position value not supported." +\
                                                          "\nValue should be bottom or top"

        if len(draw_text) > 0:

            font = ImageFont.truetype(os.path.join(self.base_folder, "resource/font/Asap-Bold.ttf"), text_size)
            canvas = ImageDraw.Draw(object_image)

            y_margin = 4
            x_margin = 8
            f_w, f_h = canvas.textsize(draw_text, font)
            im_w, im_h = object_image.size
            x = int(im_w/2)-int(f_w/2)
            if h_position == 'left':
                x = x_margin
            elif h_position == 'right':
                x = int(im_w) - int(f_w) - x_margin
            if v_position == 'bottom':
                y = im_h-f_h-y_margin
            elif v_position == 'top':
                y = 0 + y_margin
            elif v_position == 'center':
                y = int(im_h/2) - int(f_h/2)

            canvas.text((x,y), draw_text, font=font, fill=text_fill)
